{# Extends base.html and adds zoomable memory chart without sliders. #}
{% extends "base.html" %}

{% block topbar_buttons %}
{{ super() }}
<button id="memoryGraphButton" class="btn btn-outline-light mr-3" data-toggle="modal" data-target="#memoryModal" onclick="javascript:resizeMemoryGraph();">Memory Graph</button>
{% endblock %}

{% block extra_nav %}
<nav class="navbar navbar-dark bg-dark px-0">
  <div id="smallMemoryGraph" class="w-100" data-toggle="modal" data-target="#memoryModal" onclick="javascript:resizeMemoryGraph();"></div>
</nav>
{% endblock %}

{% block extra_modal %}
  <!-- Memory Modal -->
  <div class="modal fade" id="memoryModal" tabindex="-1" role="dialog" aria-labelledby="memoryModalLabel" aria-hidden="true">
    <div class="modal-dialog modal-xl" role="document">
      <div class="modal-content">
        <div class="modal-header">
          <h5 class="modal-title" id="memoryModalLabel">Resident set size over time</h5>
          <button type="button" class="close" data-dismiss="modal" aria-label="Close">
            <span aria-hidden="true">&times;</span>
          </button>
        </div>
        <div class="modal-body">
          <div id="memoryGraph"></div>
        </div>
        <div class="modal-footer">
          <button type="button" class="btn btn-primary" data-dismiss="modal">Close</button>
        </div>
      </div>
    </div>
  </div>
{% endblock %}
