/* Blocks */
.d3-flame-graph rect {
  fill-opacity: 0.8;
  stroke: #1111;
  stroke-width: 1px;
}
.d3-flame-graph .frame:hover rect {
  stroke: #1118;
  stroke-width: 1px;
  cursor: pointer;
}

.d3-flame-graph-label {
  color: black;
  pointer-events: none;

  font-size: 13px;
  line-height: 1.5;

  text-align: left;
  text-overflow: ellipsis;
  overflow: hidden;
  white-space: nowrap;

  padding: 0;
  margin-left: 4px;
  margin-right: 4px;
}

/* Bacckground stuff */
.d3-flame-graph .fade {
  opacity: 0.6 !important;
}

/* Tooltip */
.d3-flame-graph-tip {
  z-index: 10;

  background: black;
  border: none;
  color: white;

  text-align: left;

  min-width: 250px;
  border-radius: 2px;
  padding: 2px 4px;
}

.tooltip-inner {
  max-width: 300px;
}

/* Loading animation */

#loading {
  position: absolute;
  top: 50%;
  left: 50%;
  transform: translate(-50%, -50%);
  text-align: center;
}

.loading-spinner {
  border: 4px solid #f3f3f3;
  border-top: 4px solid #3498db;
  border-radius: 50%;
  width: 50px;
  height: 50px;
  animation: spin 2s linear infinite;
}

@keyframes spin {
  0% {
    transform: rotate(0deg);
  }
  100% {
    transform: rotate(360deg);
  }
}

#overlay {
  position: fixed;
  top: 0;
  left: 0;
  width: 100%;
  height: 100%;
  background-color: rgba(0, 0, 0, 0.5); /* semi-transparent black */
  z-index: 99; /* make sure it's on top of other elements */
}

.flamegraph-icon {
  display: flex;
  align-items: center;
  justify-content: center;
}

.flipped svg { transform: scale(1,-1); }
